#include "parseconfig.h"
#include <errno.h>

aws_msg create_single_msg(cfg_t *source , char *def_format) {
    aws_msg result;

    result.screen = cfg_getint(source,F_SCREEN);
    result.widget_name = strdup(cfg_getstr(source,F_WIDGET_NAME));
    result.widget_command = strdup(cfg_getstr(source,F_WIDGET_COMMAND));
    result.message = strdup(cfg_getstr(source,F_MESSAGE));
    result.sb_name = strdup(cfg_getstr(source,F_SBNAME));
    result.message_format = strdup(cfg_getstr(source,I_MSG_FORMAT));
    if (!strlen(result.message_format)) {
        free(result.message_format);
        result.message_format = strdup(def_format);
    }

    return result;
}

aws_msg *create_multi_msg(cfg_t *source , int *nr , char *def_format) {
    aws_msg *result = NULL;
    cfg_t *msg;
    int i;
    
    *nr = cfg_size(source,MSG_STR);
    result = calloc(sizeof(aws_msg),*nr);
    if (result==NULL) return result;

    for (i=0 ; i<*nr ; i++) {
        msg = cfg_getnsec(source,MSG_STR,i);
        result[i] = create_single_msg(msg,def_format);
    }
    return result;
}    



int parse_config (char const *config_file , aws_config *config) {
    cfg_opt_t opts_message [] = {
        CFG_INT(F_SCREEN,        0,    CFGF_NONE),
        CFG_STR(F_SBNAME,        "",    CFGF_NONE),
        CFG_STR(F_WIDGET_NAME,        "",    CFGF_NONE),
        CFG_STR(F_WIDGET_COMMAND,    "",    CFGF_NONE),
        CFG_STR(I_MSG_FORMAT,   "", CFGF_NONE),
        CFG_STR(F_MESSAGE,        "",    CFGF_NONE),
        CFG_END()
    };

    // single options (parts of main options)
    cfg_opt_t opts_single_disk [] = {
        CFG_STR(I_DISK_MNT_POINT, "/",CFGF_NONE),
        CFG_INT(I_BLOCKSIZE,4,CFGF_NONE),
        CFG_INT(F_PRECISION,1,CFGF_NONE),
        CFG_SEC(MSG_STR,    opts_message, CFGF_MULTI),
        CFG_END()
    };

    cfg_opt_t opts_mpd_msg [] = {
        CFG_STR(I_MPD_PAUSEMSG, "<< $A - $T >>",CFGF_NONE),
        CFG_STR(I_MPD_STOPMSG,     "mpd stopped",    CFGF_NONE),
        CFG_STR(I_MPD_UKARTIST,    "unknown artist",CFGF_NONE),
        CFG_STR(I_MPD_UKTITLE,    "unknown title",CFGF_NONE),
        CFG_STR(I_MPD_FALLBACK,    "$F" ,    CFGF_NONE),
        CFG_STR(I_MPD_NOTCONNECT, "mpd not running", CFGF_NONE),
        CFG_SEC(MSG_STR,    opts_message, CFGF_MULTI),
        CFG_END()
    };    

    cfg_opt_t opts_netif [] = {
        CFG_STR(I_NETREG,"eth0",CFGF_NONE),
        CFG_SEC(MSG_STR,opts_message,CFGF_MULTI),
        CFG_END()
    };

    cfg_opt_t opts_soundcard_mixer [] = {
        CFG_STR(I_SND_MIXER_NAME,"Master",CFGF_NONE),
        CFG_SEC(MSG_STR, opts_message, CFGF_MULTI),
        CFG_END()
    };

    cfg_opt_t opts_sensor [] = {
        CFG_STR(I_SENSOR_FILE,"",CFGF_NONE),
        CFG_INT(I_SENSOR_DIV,1,CFGF_NONE),
        CFG_SEC(MSG_STR,opts_message, CFGF_MULTI),
        CFG_END()
    };

    // main options
    cfg_opt_t opts_date [] = {
        CFG_SEC(MSG_STR,opts_message,CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };

    cfg_opt_t opts_disks [] = {
        CFG_SEC(I_SINGLE_DISK, opts_single_disk, CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    
    
    cfg_opt_t opts_mbox [] = {
        CFG_STR(I_MBOX_FILE,"~/mbox",CFGF_NONE),
        CFG_SEC(MSG_STR,    opts_message, CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };

    cfg_opt_t opts_mpd [] = {
        CFG_STR(I_MPD_HOST,    "localhost",    CFGF_NONE),
        CFG_INT(I_MPD_PORT,    6600,        CFGF_NONE),
        CFG_SEC(I_MPD_MSG_STR,    opts_mpd_msg,    CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    cfg_opt_t opts_soundcard [] = {
        CFG_SEC(I_SND_MIXER, opts_soundcard_mixer, CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    cfg_opt_t opts_cpu [] = {
        CFG_STR(I_CPU_FILESTAT,"/proc/stat",CFGF_NONE),
        CFG_STR(I_CPU_FILEINFO,"/proc/cpuinfo",CFGF_NONE),
        CFG_INT(F_PRECISION,0,CFGF_NONE),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_SEC(MSG_STR, opts_message,    CFGF_MULTI),
        CFG_END()
    };    

    cfg_opt_t opts_sensors[] = {
        CFG_SEC(I_SENSOR,opts_sensor,CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    cfg_opt_t opts_mem [] = {
        CFG_STR(I_MEMFILE, "/proc/meminfo", CFGF_NONE),
        CFG_INT(F_PRECISION,1,CFGF_NONE),
        CFG_SEC(MSG_STR,opts_message,CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    cfg_opt_t opts_net [] = {
        CFG_STR(I_NETFILE, "/proc/net/dev", CFGF_NONE),
        CFG_SEC(I_NETREGS,opts_netif,CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    cfg_opt_t opts_uptime [] = {
        CFG_STR(I_UPTIMEFILE,"/proc/uptime",CFGF_NONE),
        CFG_SEC(MSG_STR,opts_message,CFGF_MULTI),
        CFG_INT(F_UPDATE_INTERVAL, 10, CFGF_NONE),
        CFG_END()
    };    

    // the main-main section :)
    cfg_opt_t opts [] = {
        CFG_STR(I_SOCKFILE,".awesome_ctl.0",CFGF_NONE),
        CFG_STR(I_MSG_FORMAT,"widget_tell $screen $sb_name $widget_name $prop $widget_name $msg", CFGF_NONE),
        CFG_SEC(I_MPD,    opts_mpd,    CFGF_NONE    ),
        CFG_SEC(I_DATE,    opts_date,    CFGF_NONE    ),
        CFG_SEC(I_SND,    opts_soundcard,    CFGF_NONE    ),
        CFG_SEC(I_CPU,    opts_cpu,    CFGF_NONE    ),
        CFG_SEC(I_DISKS, opts_disks,    CFGF_NONE    ),
        CFG_SEC(I_SENSORS, opts_sensors, CFGF_NONE    ),
        CFG_SEC(I_MEM,    opts_mem,    CFGF_NONE    ),
        CFG_SEC(I_NET,    opts_net,    CFGF_NONE    ),
        CFG_SEC(I_UPTIME, opts_uptime,    CFGF_NONE    ),
        CFG_SEC(I_MBOX, opts_mbox,  CFGF_NONE),
        CFG_END()
    };

    cfg_t *cfg;

    cfg = cfg_init(opts,CFGF_NONE);
    int ret;
    switch ((ret=cfg_parse(cfg,config_file))) {
            case CFG_FILE_ERROR:
                printf("Parsing configuration file error: %s\n",strerror(errno));
                return 1;
                break;
            case CFG_PARSE_ERROR:
                cfg_error(cfg,"Parsing configuration file failed.\n");
                return 1;
                break;
            case CFG_SUCCESS:    
                break;
    }
                

    // general configs
    config->sockfile = strdup(cfg_getstr(cfg,I_SOCKFILE));
    config->message_format = strdup(cfg_getstr(cfg,I_MSG_FORMAT));

    // sensor's configuration
    cfg_t *cfg_sensors;
    cfg_sensors = cfg_getsec(cfg,I_SENSORS);
    if (cfg_sensors!=NULL) {
        cfg_t *cfg_sensor;
        int i;

        config->sensor.sensor_nr = cfg_size(cfg_sensors,I_SENSOR);
        config->sensor.sensors = calloc(sizeof(aws_single_sensor),config->sensor.sensor_nr);
        config->sensor.update_interval = cfg_getint(cfg_sensors,F_UPDATE_INTERVAL);
        for (i=0 ; i<config->sensor.sensor_nr ; i++) {
            cfg_sensor = cfg_getnsec(cfg_sensors,I_SENSOR,i);
            config->sensor.sensors[i].file = strdup(cfg_getstr(cfg_sensor,I_SENSOR_FILE));
            config->sensor.sensors[i].div = cfg_getint(cfg_sensor,I_SENSOR_DIV);
            if (!config->sensor.sensors[i].div) config->sensor.sensors[i].div=1;
            config->sensor.sensors[i].msg = create_multi_msg(cfg_sensor,&(config->sensor.sensors[i].msg_nr),config->message_format);
        }    
    }
    
    // network's configuration
    cfg_t *cfg_net;
    cfg_net = cfg_getsec(cfg,I_NET);
    if (cfg_net != NULL) {
        cfg_t *cfg_ifs;
        int i;

        config->net.file = strdup(cfg_getstr(cfg_net,I_NETFILE));
        config->net.if_nr = cfg_size(cfg_net,I_NETREGS);
        config->net.ifs=calloc(sizeof(aws_if),config->net.if_nr);
        config->net.update_interval = cfg_getint(cfg_net,F_UPDATE_INTERVAL);
        for (i=0 ; i<config->net.if_nr ; i++) {
            cfg_ifs = cfg_getnsec(cfg_net,I_NETREGS,i);
            config->net.ifs[i].nif = strdup(cfg_getstr(cfg_ifs,I_NETREG));
            config->net.ifs[i].msg = create_multi_msg(cfg_ifs,&(config->net.ifs[i].msg_nr),config->message_format);
        }    
    }    

    // cpu's configuration
    cfg_t *cfg_cpu;
    cfg_cpu = cfg_getsec(cfg,I_CPU);
    if (cfg_cpu!=NULL) {
        config->cpu.file_stat = strdup(cfg_getstr(cfg_cpu,I_CPU_FILESTAT));
        config->cpu.file_info = strdup(cfg_getstr(cfg_cpu,I_CPU_FILEINFO));
        config->cpu.precision = cfg_getint(cfg_cpu,F_PRECISION);
        config->cpu.update_interval = cfg_getint(cfg_cpu,F_UPDATE_INTERVAL);
        config->cpu.msg  = create_multi_msg(cfg_cpu,&(config->cpu.msg_nr),config->message_format);
    }

    // date/time's configuration
    cfg_t *cfg_date;
    cfg_date = cfg_getsec(cfg,I_DATE);
    if (cfg_date!=NULL) {
        config->date.update_interval = cfg_getint(cfg_date,F_UPDATE_INTERVAL);
        config->date.msg    = create_multi_msg(cfg_date,&(config->date.msg_nr),config->message_format);    
    }

    // sound card's configuration
    cfg_t *cfg_sound;
    cfg_sound = cfg_getsec(cfg,I_SND);
    if (cfg_sound!=NULL) {
        int i; 
        cfg_t *cfg_sound_mixer;

        config->sound.update_interval = cfg_getint(cfg_sound,F_UPDATE_INTERVAL);
        config->sound.mixer_nr = cfg_size(cfg_sound,I_SND_MIXER);
        config->sound.mixers=calloc(sizeof(aws_snd_mixer),config->sound.mixer_nr);
        for ( i=0 ; i<config->sound.mixer_nr ; i++ ) {
            cfg_sound_mixer = cfg_getnsec(cfg_sound,I_SND_MIXER,i);
            config->sound.mixers[i].mixer = strdup(cfg_getstr(cfg_sound_mixer,I_SND_MIXER_NAME));    
            config->sound.mixers[i].msg = create_multi_msg(cfg_sound_mixer,&(config->sound.mixers[i].msg_nr),config->message_format);
        }    
    }    
    
    // memory's configuration
    cfg_t *cfg_mem;
    cfg_mem = cfg_getsec(cfg,I_MEM);
    if (cfg_mem!=NULL) {
        config->mem.file = strdup(cfg_getstr(cfg_mem,I_MEMFILE));
        config->mem.precision = cfg_getint(cfg_mem,F_PRECISION);
        config->mem.update_interval = cfg_getint(cfg_mem,F_UPDATE_INTERVAL);
        config->mem.msg = create_multi_msg(cfg_mem,&(config->mem.msg_nr),config->message_format);
    }    

    // mbox's configuration
    cfg_t *cfg_mbox;
    cfg_mbox = cfg_getsec(cfg,I_MBOX);
    if (cfg_mbox!=NULL) {
        config->mbox.file = strdup(cfg_getstr(cfg_mbox,I_MBOX_FILE));
        config->mbox.update_interval = cfg_getint(cfg_mbox,F_UPDATE_INTERVAL);
        config->mbox.msg = create_multi_msg(cfg_mbox,&(config->mbox.msg_nr),config->message_format);
    }    

    // mpd's configuration    
    cfg_t *cfg_mpd;
    cfg_mpd = cfg_getsec(cfg,I_MPD);
    if (cfg_mpd!=NULL) {
        config->mpd.port    = cfg_getint(cfg_mpd,I_MPD_PORT);
        config->mpd.host    = strdup(cfg_getstr(cfg_mpd,I_MPD_HOST));
        config->mpd.update_interval = cfg_getint(cfg_mpd,F_UPDATE_INTERVAL);
        config->mpd.msg_nr    = cfg_size(cfg_mpd,I_MPD_MSG_STR);
        config->mpd.msg        = calloc(sizeof(aws_mpd_msg),config->mpd.msg_nr);
        int j; cfg_t *cfg_mpd_msg;
        for (j=0 ; j<config->mpd.msg_nr ; j++) {
            cfg_mpd_msg = cfg_getnsec(cfg_mpd,I_MPD_MSG_STR,j);
            config->mpd.msg[j].msg_pause = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_PAUSEMSG));
            config->mpd.msg[j].msg_stop  = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_STOPMSG));
            config->mpd.msg[j].msg_ukartist = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_UKARTIST));
            config->mpd.msg[j].msg_uktitle    = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_UKTITLE));
            config->mpd.msg[j].msg_fallback    = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_FALLBACK));
            config->mpd.msg[j].msg_notconnect = strdup(cfg_getstr(cfg_mpd_msg,I_MPD_NOTCONNECT));
            config->mpd.msg[j].msg    = create_multi_msg(cfg_mpd_msg,&(config->mpd.msg[j].msg_nr),config->message_format);
        }
    }

    // disks configuration
    cfg_t *cfg_disks;
    cfg_disks = cfg_getsec(cfg,I_DISKS);
    if (cfg_disks!=NULL) {
        config->disks.disk_nr    = cfg_size(cfg_disks,I_SINGLE_DISK);
        config->disks.disks    = calloc(sizeof(aws_single_disk),config->disks.disk_nr);
        config->disks.update_interval = cfg_getint(cfg_disks,F_UPDATE_INTERVAL);
        int j; cfg_t *cfg_single_disk;
        for (j=0 ; j<config->disks.disk_nr ; j++) {
            cfg_single_disk = cfg_getnsec(cfg_disks,I_SINGLE_DISK,j);
            config->disks.disks[j].mnt_point = strdup(cfg_getstr(cfg_single_disk,I_DISK_MNT_POINT));
            config->disks.disks[j].blocksize = cfg_getint(cfg_single_disk,I_BLOCKSIZE);
            config->disks.disks[j].precision = cfg_getint(cfg_single_disk,F_PRECISION);
            config->disks.disks[j].msg = create_multi_msg(cfg_single_disk,&(config->disks.disks[j].msg_nr),config->message_format);
        }
    }    

    // uptime's configuration
    cfg_t *cfg_uptime;
    cfg_uptime = cfg_getsec(cfg,I_UPTIME);
    if (cfg_uptime!=NULL) {
        config->uptime.update_interval = cfg_getint(cfg_uptime,F_UPDATE_INTERVAL);
        config->uptime.file = strdup(cfg_getstr(cfg_uptime,I_UPTIMEFILE));
        config->uptime.msg  = create_multi_msg(cfg_uptime,&(config->uptime.msg_nr),config->message_format);
    }    
    cfg_free(cfg);
    return 0;
}

void printmsg (const char* prefix , aws_msg msg) {
    printf("%sMessage format: %s\n\
            %sScreen: %d\n\
            %sStatusbar name: %s\n\
            %sWidget name: %s\n\
            %sWidget property: %s\n\
            %sMessage: %s\n",
                prefix, msg.message_format,
                prefix, msg.screen,
                prefix, msg.sb_name,
                prefix, msg.widget_name,
                prefix, msg.widget_command,
                prefix, msg.message );
}

void printinfo (aws_config myconfig) {
    int i,j;

        printf("CPU\n");
        printf("\tInterval: %d\n",myconfig.cpu.update_interval);
        printf("\t%s\n",myconfig.cpu.file_stat);
        printf("\t%s\n",myconfig.cpu.file_info);
        for ( i=0 ; i<myconfig.cpu.msg_nr ; i++) {
            printmsg("\t",myconfig.cpu.msg[i]);
        }    

        printf("Date\n");
        printf("\tInterval: %d\n",myconfig.date.update_interval);
        for ( i=0 ; i<myconfig.date.msg_nr ; i++) {
            printmsg("\t",myconfig.date.msg[i]);
        }

        printf("Disks\n");
        printf("\tInterval: %d\n",myconfig.disks.update_interval);
        for ( i=0 ; i<myconfig.disks.disk_nr ; i++) {
            printf("\tMount point: %s Blocksize: %d Precision: %d\n",myconfig.disks.disks[i].mnt_point,myconfig.disks.disks[i].blocksize,myconfig.disks.disks[i].precision);
            for ( j=0 ; j<myconfig.disks.disks[i].msg_nr ; j++) {
                printmsg("\t\t",myconfig.disks.disks[i].msg[j]);
            }
        }    

        printf("Mbox\n");
        printf("\tInterval: %d\n",myconfig.mbox.update_interval);
        printf("\tFile: %s\n",myconfig.mbox.file);
        for ( i=0 ; i<myconfig.mbox.msg_nr ; i++) {
            printmsg("\t",myconfig.mbox.msg[i]);
        }

        printf("Mem\n");
        printf("\tInterval: %d\n",myconfig.mem.update_interval);
        printf("\tFile: %s\n",myconfig.mem.file);
        for ( i=0 ; i<myconfig.mem.msg_nr ; i++ ) {
            printmsg("\t",myconfig.mem.msg[i]);
        }    

        printf("MPD\n");
        printf("\tInterval: %d\n",myconfig.mpd.update_interval);
        printf("\thost: %s  port: %d\n",myconfig.mpd.host,myconfig.mpd.port);
        for ( i=0 ; i<myconfig.mpd.msg_nr ; i++) {
            printf("\tpause: %s\n\tstop: %s\n\tunknown artist: %s\n\tunknown title: %s\n\tfallback msg: %s\n\tnot connect: %s\n",
                myconfig.mpd.msg[i].msg_pause,
                myconfig.mpd.msg[i].msg_stop,
                myconfig.mpd.msg[i].msg_ukartist,
                myconfig.mpd.msg[i].msg_uktitle,
                myconfig.mpd.msg[i].msg_fallback,
                myconfig.mpd.msg[i].msg_notconnect);
            for (j=0 ; j<myconfig.mpd.msg[i].msg_nr ; j++) {
                printmsg("\t",myconfig.mpd.msg[i].msg[j]);
            }    
        }

        printf("Net\n");
        printf("\tInterval: %d\n",myconfig.net.update_interval);
        printf("\tFile: %s\n",myconfig.net.file);
        for ( i=0 ; i<myconfig.net.if_nr ; i++ ) {
            printf("\tIf: %s\n",myconfig.net.ifs[i].nif);
            for (j=0 ; j<myconfig.net.ifs[i].msg_nr ; j++) printmsg("\t",myconfig.net.ifs[i].msg[j]);
        }

        printf("Sensors\n");
        printf("\tInterval: %d\n",myconfig.sensor.update_interval);
        for ( i=0 ; i<myconfig.sensor.sensor_nr ; i++ ) {
            printf("\tFile: %s\n",myconfig.sensor.sensors[i].file);
            for (j=0 ; j<myconfig.sensor.sensors[i].msg_nr ; j++) {
                printmsg("\t\t",myconfig.sensor.sensors[i].msg[j]);
            }
        }

        printf("Soundcard\n");
        printf("\tInterval: %d\n",myconfig.sound.update_interval);
        for ( i=0 ; i<myconfig.sound.mixer_nr ; i++) {
            printf("\t%s\n",myconfig.sound.mixers[i].mixer);
            for (j=0 ; j<myconfig.sound.mixers[i].msg_nr ; j++) {
                printmsg("\t\t",myconfig.sound.mixers[i].msg[j]);
            }
        }    

        printf("Uptime\n");
        printf("\tInterval: %d\n",myconfig.uptime.update_interval);
        for ( i=0 ; i<myconfig.uptime.msg_nr ; i++) {
            printmsg("\t",myconfig.uptime.msg[i]);
        }    
}    

// vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:encoding=utf-8:textwidth=80
