#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <confuse.h>

#define MSG_STR "msg"

// flags
#define F_SCREEN "screen"
#define F_WIDGET_COMMAND "widget_property"
#define F_WIDGET_NAME "widget_name"
#define F_MESSAGE "message"
#define F_SBNAME "sb_name"
#define F_UPDATE_INTERVAL "interval"
#define F_PRECISION "precision"

// identifiers
#define I_SOCKFILE "sockfile"
#define I_MSG_FORMAT "message_format"

#define I_DATE "date"

#define I_DISKS "disks"
#define I_SINGLE_DISK "disk"
#define I_BLOCKSIZE "blocksize"
#define I_DISK_MNT_POINT "mnt_point"

#define I_SND "snd"
#define I_SND_MIXER "mixer"
#define I_SND_MIXER_NAME "mixer_name"

#define I_MPD "mpd"
#define I_MPD_HOST "host"
#define I_MPD_PORT "port"
#define I_MPD_MSG_STR "mpd_msg"
#define I_MPD_PAUSEMSG "pause_msg"
#define I_MPD_STOPMSG "stop_msg"
#define I_MPD_UKARTIST "uk_artist" //UnKnown artist
#define I_MPD_UKTITLE "uk_title"
#define I_MPD_FALLBACK "fallback"
#define I_MPD_NOTCONNECT "notconnect"

#define I_MBOX "mbox"
#define I_MBOX_FILE "file"

#define I_CPU "cpu"
#define I_CPU_FILESTAT "file_stat"
#define I_CPU_FILEINFO "file_info"

#define I_SENSOR "sensor"
#define I_SENSORS "sensors"
#define I_SENSOR_FILE "file"
#define I_SENSOR_DIV "div"

#define I_MEM "mem"
#define I_MEMFILE "file"

#define I_NET "net"
#define I_NETFILE "file"
#define I_NETREGS "ifs"
#define I_NETREG "nif"

#define I_UPTIME "uptime"
#define I_UPTIMEFILE "file"

// some structure definitions
typedef struct aws_msg {
    int screen;
    char *widget_name;
    char *sb_name;
    char *widget_command; // 'text' or 'data <progressbar_data_name>' or similar
    char *message_format;
    char *message;
} aws_msg;

typedef struct aws_snd_mixer {
    char *mixer;
    int msg_nr;
    aws_msg *msg;
} aws_snd_mixer;

typedef struct aws_mpd_msg {
    char
        *msg_pause, 
        *msg_stop, 
        *msg_ukartist, 
        *msg_uktitle,
        *msg_fallback,
        *msg_notconnect;
    int update_interval;
    int msg_nr;
    aws_msg *msg;
} aws_mpd_msg;

// individual config structures
typedef struct aws_cpu {
    char *file_stat , *file_info;
    int precision;
    int update_interval;
    int msg_nr;
    aws_msg *msg;
} aws_cpu;    

typedef struct aws_date {
    int update_interval;
    int msg_nr;    
    aws_msg *msg;
} aws_date;    

typedef struct aws_single_disk {
    char *mnt_point;
    int blocksize;
    int msg_nr;
    int precision; // number of decimals
    aws_msg *msg;
} aws_single_disk;

typedef struct aws_disks {
    int update_interval;
    int disk_nr;
    aws_single_disk *disks;
} aws_disks;    

typedef struct aws_mbox {
    char *file;
    int update_interval;
    int msg_nr;
    aws_msg *msg;
} aws_mbox;    

typedef struct aws_mem {
    char *file;
    int precision;
    int update_interval;
    int msg_nr;
    aws_msg *msg;
} aws_mem;    

typedef struct aws_moc {
    char *sockfile;
} aws_moc;

typedef struct aws_mpd {
    char *host;
    int port;
    int update_interval;
    int msg_nr;
    aws_mpd_msg *msg;
} aws_mpd;    

typedef struct aws_single_sensor {
    char *file;
    int div;
    int msg_nr;
    aws_msg *msg;
} aws_single_sensor;    

typedef struct aws_sensor {
    int update_interval;
    int sensor_nr;
    aws_single_sensor *sensors;
} aws_sensor;    

typedef struct aws_sound {
    int update_interval;
    int mixer_nr;
    aws_snd_mixer *mixers;
} aws_sound;    

typedef struct aws_if {
    char *nif; // Net InterFace
    int msg_nr;
    aws_msg *msg;
} aws_if;

typedef struct aws_net {
    char *file;
    int update_interval;
    int if_nr;
    aws_if *ifs;
} aws_net;    

typedef struct aws_uptime {
    char *file;
    int update_interval;
    int msg_nr;
    aws_msg *msg;
} aws_uptime;

// the main config structure
typedef struct aws_config {
    char *sockfile;
    char *message_format;
    aws_cpu     cpu;
    aws_date    date;
    aws_disks   disks;
    aws_mbox    mbox;
    aws_mem     mem;
//    aws_moc        moc;
    aws_mpd     mpd;
    aws_net     net;
    aws_sensor  sensor;
    aws_sound   sound;
    aws_uptime  uptime;
} aws_config;

// interface
int parse_config (const char *config_file , aws_config *config); 
void printinfo(aws_config);

// vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:encoding=utf-8:textwidth=8
